/*

  phalic 1.1 - july 2000
  by mirage / KaotiK Team <mirage@kaotik.org>

  you can start phalic from the command line or from inetd.
  example inetd.conf line:
  69	stream	tcp	nowait	root	/usr/sbin/tcpd /usr/bin/phalic -u nobody -g nobody -f /etc/phalic.txt -i

  the ascii was ripped from some server to which i don't have access nor
   source code to their daemon. as such, i coded my own...
  in the text file, each frame is separated by a line containing only a \001
   (ascii 1), to apply a (configurable) delay between frames.

*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <pwd.h>

#define BUF_SIZE 10000
#define HOST_SIZE 64
#define DELAY 50000
#define LISTEN_PORT 6969
#define FILENAME "phalic.txt"

int listen_sock;
int listen_port = LISTEN_PORT;
int delay = DELAY;
char filename[BUF_SIZE];
char src[HOST_SIZE+1];

void
halt (const char *error)
{
  perror (error);
  exit (1);
}

void
draw (void)
{
  char buf[BUF_SIZE];
  FILE *f = fopen (filename, "r");
  if (f == NULL)
    halt ("fopen()");
  while (fgets (buf, BUF_SIZE, f) != NULL)
    {
      buf[strlen (buf)-1] = '\0';
      if (strcmp (buf, "\001") != 0)
        {
          printf ("%s\r\n", buf);
          fflush (stdout);
        }
      else
        usleep (delay);
    }
  fclose (f);
  exit (0);
}

void
setup_socket (void)
{
  int i = 1;
  struct sockaddr_in local;
  listen_sock = socket (AF_INET, SOCK_STREAM, 0);
  if (listen_sock == -1)
    halt ("socket()");
  setsockopt (listen_sock, SOL_SOCKET, SO_REUSEADDR, (void *)&i, sizeof (int));
  memset (&local, 0, sizeof (local));
  local.sin_family = AF_INET;
  local.sin_port = htons (listen_port);
  if (src[0] == 0)
    local.sin_addr.s_addr = INADDR_ANY;
  else
    {
      struct hostent *he;
      he = gethostbyname (src);
      if (he == NULL)
        halt ("gethostbyname()");
      memcpy (&local.sin_addr, he->h_addr, he->h_length);
    }
  memset (&(local.sin_zero), 0, 8);
  if (bind (listen_sock, (struct sockaddr *)&local, sizeof (struct sockaddr)) == -1)
    halt ("bind()");
  if (listen (listen_sock, 5) == -1)
    halt ("listen()");
}

void
accept_socket (void)
{
  int fd, i = sizeof (struct sockaddr_in);
  struct sockaddr_in remote;
  while (1)
    {
      fd = accept (listen_sock, (struct sockaddr *)&remote, &i);
      if (fd == -1)
        exit (1);
      if (!fork ())
        {
          dup2 (fd, 1);
          dup2 (fd, 2);
          draw ();
        }
      close (fd);
    }
}

void
bg (void)
{
  int fd = open ("/dev/tty", O_RDWR);
  if (fd != -1)
    {
      ioctl (fd, TIOCNOTTY, 0);
      close (fd);
    }
  switch (fork ())
    {
      case 0:
        break;
      case -1:
        halt ("fork()");
      default:
        exit (0);
    }
  signal (SIGCHLD, SIG_IGN);
}

int
main (int argc, char *argv[])
{
  int i, id, inetd = 0, help = 0;
  snprintf (filename, BUF_SIZE, "%s", FILENAME);
  src[0] = 0;

  while ((i = getopt (argc, argv, "p:f:d:b:u:g:ih")) != EOF)
    switch (i)
      {
        case 'p':
          listen_port = atoi (optarg);
          break;
        case 'f':
          snprintf (filename, BUF_SIZE, "%s", optarg);
          break;
        case 'd':
          delay = atoi (optarg);
          break;
        case 'b':
          snprintf (src, HOST_SIZE, "%s", optarg);
          break;
        case 'u':
          if (optarg[0] >= '0' && optarg[0] <= '9')
            id = atoi (optarg);
          else
            {
              struct passwd *pwd = getpwnam (optarg);
              if (pwd == NULL)
                halt ("getpwnam()");
              id = pwd->pw_uid;
            }
          if (setuid (id) == -1)
            halt ("setuid()");
          break;
        case 'g':
          if (setgid (atoi (optarg)) == -1)
            halt ("setgid()");
          break;
        case 'i':
          inetd = 1;
          break;
        case 'h':
          help = 1;
          break;
        default:
          return 1;
      }

  if (inetd)
    draw ();

  printf ("phalic 1.1, by mirage / KaotiK Team <mirage@kaotik.org> (-h for help)\n");

  if (help)
    {
      printf ("usage: %s [options]\n", argv[0]);
      printf ("  -p <port> port to listen to (default %d)\n", LISTEN_PORT);
      printf ("  -f <filename> file to show (default %s)\n", FILENAME);
      printf ("  -d <delay> delay in microseconds between frames (default %d)\n", DELAY);
      printf ("  -b <address> bind socket to this local address (default any)\n");
      printf ("  -u <uid|username> set user id\n");
      printf ("  -g <gid> set group id\n");
      printf ("  -i inetd mode\n");
      return 0;
    }

  printf ("listening on port %d, using file %s\n", listen_port, filename);
  setup_socket ();
  bg ();
  accept_socket ();

  return 0;
}

